__version__ = "1.0.0"

from . import defaults
from . import exceptions
from . import main
from .parser import ParserSettings
from . import path
from . import templates
from .templates import RawTemplate
from .templates import CogTemplate
from .templates import JinjaTemplate
from . import utils

import os



"""
Visits directories and executes python scripts to perform content generation

Usage
-----
autojinja [OPTIONS] (PYTHON_SCRIPT | DIRECTORY)...

Examples
--------
autojinja script1.py script2.py
autojinja -a .

Arguments
---------
PYTHON_SCRIPT:
    Python script to execute
    The filepath must end with '.py' extension

DIRECTORY:
    Directory to visit
    Only works if '--search-filename' or '--search-tag' is enabled

OPTIONS:
    -f, --search-filename         Searches for python scripts named '__jinja__.py' in visited directories
    -t, --search-tag              Searches for python scripts tagged 'autojinja' in visited directories
    -r, --recursive               Recursively visits subdirectories of visited directories
    -a, --all                     All above options simultaneously (equivalent to '-ftr')
    --filename=FILENAME           Filename searched by '--search-filename'. Default value is '__jinja__.py'
    --tag=TAG                     Tag searched by '--search-tag'. Default value is 'autojinja'
                                  Python scripts' first line must contain this tag (ex: '### autojinja ###')
    -e, --env=NAME=VALUE/FILE     Additional environment variable or .env file
    -i, --includes=DIRECTORIES    Additional import directories for executed python scripts
                                  Directory list separated by ':' (Unix only) or ';' (Windows and Unix)
                                  Appended to environment variable 'PYTHONPATH'
   --remove-markers=ENABLE        Removes markers from generated outputs
                                  Default value is '0' or environment variable 'AUTOJINJA_REMOVE_MARKERS'
                                  Overrides environment variable 'AUTOJINJA_REMOVE_MARKERS'
   --silent                       Prevents executed python scripts from writing to stdout/stderr
                                  Enabled if environment variable 'AUTOJINJA_SILENT' == 1
                                  Overrides environment variable 'AUTOJINJA_SILENT'
"""

from .defaults import *
from . import path
from . import utils

import argparse
import os.path
import subprocess
import sys

this_module = sys.modules[__name__]

def main(*arguments):
    ### Parse arguments
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                     description="Visits directories and executes python scripts to perform content generation")
    parser.add_argument("arguments",
                        nargs="+",
                        help="PYTHON_SCRIPT:\n"
                             "    Python script to execute\n"
                             "    The filepath must end with '.py' extension\n\n"
                             "DIRECTORY:\n"
                             "    Directory to visit\n"
                             "    Only works if '--search-filename' or '--search-tag' is enabled")
    parser.add_argument("-f",
                        "--search-filename",
                        action="store_true",
                        help=f"searches for python scripts named '{AUTOJINJA_DEFAULT_FILENAME}' in visited directories")
    parser.add_argument("-t",
                        "--search-tag",
                        action="store_true",
                        help=f"searches for python scripts tagged '{AUTOJINJA_DEFAULT_TAG}' in visited directories")
    parser.add_argument("-r",
                        "--recursive",
                        action="store_true",
                        help="recursively visits subdirectories of visited directories")
    parser.add_argument("-a",
                        "--all",
                        action="store_true",
                        help="all above options simultaneously (equivalent to '-ftr')")
    parser.add_argument("--filename",
                        default=AUTOJINJA_DEFAULT_FILENAME,
                        help=f"filename searched by '--search-filename'. Default filename is '{AUTOJINJA_DEFAULT_FILENAME}'")
    parser.add_argument("--tag",
                        default=AUTOJINJA_DEFAULT_TAG,
                        help=f"tag searched by '--search-tag'. Default tag is '{AUTOJINJA_DEFAULT_TAG}'\n"
                             f"Python scripts' first line must contain this tag (ex: '### {AUTOJINJA_DEFAULT_TAG} ###')")
    parser.add_argument("-e",
                        "--env",
                        action="append",
                        help="additional environment variable or .env file")
    parser.add_argument("-i",
                        "--includes",
                        action="append",
                        help="additional import directories for executed python scripts\n"
                             "Directory list separated by ':' (Unix only) or ';' (Windows and Unix)\n"
                             "Appended to environment variable 'PYTHONPATH'")
    parser.add_argument("--remove-markers",
                        help=f"removes markers from generated outputs\n"
                             f"Default value is '0' or environment variable '{AUTOJINJA_REMOVE_MARKERS}'\n"
                             f"Overrides environment variable '{AUTOJINJA_REMOVE_MARKERS}'")
    parser.add_argument("--silent",
                        action="store_true",
                        help=f"prevents executed python scripts from writing to stdout/stderr\n"
                             f"Enabled if environment variable '{AUTOJINJA_SILENT}' == 1\n"
                             f"Overrides environment variable '{AUTOJINJA_SILENT}'")

    args = parser.parse_args(arguments)

    ### Verify options
    if args.all:
        args.search_filename = True
        args.search_tag = True
        args.recursive = True
    # Remove markers
    if args.remove_markers != None:
        if not args.remove_markers.isdigit() or int(args.remove_markers) < 0 or int(args.remove_markers) > 1:
            raise Exception("Expected 0 or 1 for '--remove-markers'")
    # Silent
    if args.silent:
        args.silent = 1

    ### Parse arguments
    def file_tagged(script):
        try:
            return utils.file_tagged(script, args.tag)
        except:
            sys.stderr.write(f"[autojinja]  Couldn't read file at path  {script.abspath}\n")

    files = []
    for x in [path(x) for x in args.arguments]:
        if x.isfile:
            if x.ext != ".py":
                raise Exception(f"File at path \"{x.abspath}\" is not a python script (.py extension)")
            files.append(x)
        elif not x.isdir:
            raise Exception(f"File or directory at path \"{x.abspath}\" doesn't exist")
        else:
            if not args.search_filename and not args.search_tag:
                raise Exception("Directory arguments require '--search-filename' or '--search-tag'")
            if args.recursive:
                for wroot, _, wfiles in os.walk(x):
                    for script in [path(wroot).join(x) for x in wfiles]:
                        ## JINJA
                        if args.search_filename:
                            if script.filename == args.filename:
                                files.append(script)
                        ## PYTHON
                        if args.search_tag:
                            if script.ext == ".py":
                                if file_tagged(script):
                                    files.append(script)
            else:
                ## JINJA
                if args.search_filename:
                    script = x.join(args.filename)
                    if script.isfile:
                        files.append(script)
                ## PYTHON
                if args.search_tag:
                    for script in x.files("*.py"):
                        if file_tagged(script):
                            files.append(script)

    # Make absolute and remove duplicates
    files = [x.abspath for x in files]
    files = list(dict.fromkeys(files))

    ### Prepare environment
    env = os.environ.copy()

    # Additional environment variables
    if args.env:
        utils.parse_envvars(env, args.env)

    # Additional import directories
    if args.includes:
        includes = [path[include].abspath for includes in args.includes for include in includes.split(utils.os_pathsep(includes))]
        if "PYTHONPATH" in env:
            env["PYTHONPATH"] = f"{env['PYTHONPATH']}{os.pathsep}{os.pathsep.join(includes)}"
        else:
            env["PYTHONPATH"] = os.pathsep.join(includes)

    # Additional options
    if args.remove_markers == None:
        if not AUTOJINJA_REMOVE_MARKERS in env:
            args.remove_markers = "0"
        else:
            args.remove_markers = env[AUTOJINJA_REMOVE_MARKERS]
            if not args.remove_markers.isdigit() or int(args.remove_markers) < 0 or int(args.remove_markers) > 1:
                raise Exception(f"Expected 0 or 1 for environment variable '{AUTOJINJA_REMOVE_MARKERS}'")
    env[AUTOJINJA_REMOVE_MARKERS] = args.remove_markers

    if not args.silent:
        if not AUTOJINJA_SILENT in env:
            args.silent = 0
        else:
            args.silent = env[AUTOJINJA_SILENT]
            if not args.silent.isdigit() or int(args.silent) < 0 or int(args.silent) > 1:
                raise Exception(f"Expected 0 or 1 for environment variable '{AUTOJINJA_SILENT}'")
            args.silent = int(args.silent)
    env[AUTOJINJA_SILENT] = "1" if args.silent else "0"

    ### Execute python scripts
    for script in files:
        process = subprocess.Popen([sys.executable, "-u", script],
                                    cwd=script.dirpath,
                                    env=env,
                                    stdout = subprocess.PIPE if args.silent else None,
                                    stderr = subprocess.STDOUT if args.silent else None,
                                    universal_newlines = True if args.silent else None)
        out, _ = process.communicate()
        errcode = process.returncode
        if errcode != 0:
            raise Exception(f"{out if args.silent else ''}Error {errcode} while executing script at path \"{script}\"")

class module_call:
    """ Overrides main() and main.attr
    """
    def __call__(self, *args, **kwargs):
        """ Calls the main function
        """
        return main(*args, **kwargs)
    def __getattribute__(self, attr):
        """ Forwards this module
        """
        return this_module.__getattribute__(attr)

sys.modules[__name__] = module_call()



from . import defaults
from . import parser
from . import path

import os
import sys

def file_tagged(filepath, tag = defaults.AUTOJINJA_DEFAULT_TAG, encoding = None):
    """ Returns True if the file at the given filepath is tagged with the given tag.
        The file's first line must contain this tag (ex: '### autojinja ###').
        Raises an error if the file can't be read.
    """
    with open(filepath, 'r', encoding = encoding or "utf-8") as file:
        return tag in file.readline()

def generate_file(filepath, new_content, old_content = None, encoding = None, newline = None):
    """ Generates the given content to the given filepath.
        Only writes the content to the file if the content is new.
        The previous content can be directly provided to avoid reading the file.
        Raises an error if the file can't be read/write.
    """
    assert filepath != None, "output filepath parameter can't be None"
    ### Compare old content
    filepath = path(filepath).abspath
    created = not filepath.isfile
    if not created and not old_content:
        with open(filepath, 'r', encoding = encoding or "utf-8") as file:
            old_content = file.read()
    ### Save new generation
    changed = new_content != old_content
    if created or changed:
        with open(filepath, 'w', encoding = encoding or "utf-8", newline = newline) as file:
            file.write(new_content)
    print(f"[autojinja]  {'  new  ' if created else 'changed' if changed else '-------'}  {filepath}  (from {path.no_antislash(sys.argv[0])})")

def parse_file(filepath, settings = None, encoding = None):
    """ Parses the given file and return the parsing result.
        Returns an empty dictionary if the file doesn't exist.
        Raises an error if the file can't be read.
    """
    if not os.path.isfile(filepath):
        return {}
    with open(filepath, 'r', encoding = encoding or (settings.encoding if settings else None) or "utf-8") as file:
        return parse_string(file.read(), settings)

def parse_string(string, settings = None):
    """ Parses the given string and returns the parsing result.
    """
    try:
        object = parser.Parser(string, settings or parser.ParserSettings())
        object.parse()
        return object
    except BaseException as e:
        raise e.with_traceback(None)

def edit_markers_from_file(filepath, settings = None, encoding = None):
    """ Returns a dictionary with all edit markers in the given file.
        Returns an empty dictionary if the file doesn't exist.
        Raises an error if the file can't be read.
    """
    object = parse_file(filepath, settings, encoding)
    return object.edit_markers

def edit_markers_from_string(string, settings = None):
    """ Returns a dictionary with all edit markers in the given string.
    """
    object = parse_string(string, settings)
    return object.edit_markers

def edits_from_file(filepath, settings = None, encoding = None):
    """ Returns a dictionary with all manual edits in the given file.
        Returns an empty dictionary if the file doesn't exist.
        Raises an error if the file can't be read.
    """
    object = parse_file(filepath, settings, encoding)
    return object.edits

def edits_from_string(string, settings = None):
    """ Returns a dictionary with all manual edits in the given string.
    """
    object = parse_string(string, settings)
    return object.edits

def parse_envvars(env, values):
    """ Loads the given environment variables to the given environment dictionary.
        Variable format must be 'name=value', otherwise considered as an environment file.
    """
    for arg in values:
        if '=' in arg: # Environment variable
            splits = arg.split('=', 1)
            env[splits[0].strip()] = evaluate_env(env, splits[1].strip())
        else: # Environment file
            parse_envfile(env, arg)

def parse_envfile(env, envfile):
    """ Loads the given environment file and appends all environment variables to the given environment dictionary.
        The special environment variable ${THIS_DIRPATH} can be used to refer to the environment file location.
    """
    with open(envfile, encoding = "utf-8") as file:
        lines = [line.split('#', 1)[0].strip() for line in file.readlines()] # Remove comments
        lines = [line for line in lines if line] # Remove empty lines
        for line in lines:
            splits = line.split('=', 1)
            name = splits[0].strip()
            if len(splits) == 1:
                env[name] = "" # No value defined
            else:
                exists = defaults.AUTOJINJA_THIS_DIRPATH in env
                if not exists:
                    env[defaults.AUTOJINJA_THIS_DIRPATH] = path(envfile).abspath.dirpath[:-1]
                value = evaluate_env(env, splits[1].strip())
                env[name] = value.strip()
                if not exists:
                    del env[defaults.AUTOJINJA_THIS_DIRPATH]

def evaluate_env(env, value):
    """ Evaluates the given string with the appropriate environment variables' value.
        /dir1/${VAR1}/file.txt -> /dir1/dir2/file.txt
    """
    old_env = os.environ
    try:
        os.environ = env
        return os.path.expandvars(value)
    finally:
        os.environ = old_env

def os_pathsep(value):
    """ Returns the appropriate path separator based the given value and the host OS.
    """
    if os.name != "nt": # Unix
        return ':' if ':' in value else ';'
    else: # Windows
        return ';'



AUTOJINJA_DEFAULT_FILENAME   = "__jinja__.py"
AUTOJINJA_DEFAULT_TAG        = "autojinja"
AUTOJINJA_DEFAULT_COG_OPEN   = "[[["
AUTOJINJA_DEFAULT_COG_CLOSE  = "]]]"
AUTOJINJA_DEFAULT_COG_END    = "end"
AUTOJINJA_DEFAULT_EDIT_OPEN  = "<<["
AUTOJINJA_DEFAULT_EDIT_CLOSE = "]>>"
AUTOJINJA_DEFAULT_EDIT_END   = "end"

AUTOJINJA_REMOVE_MARKERS = "AUTOJINJA_REMOVE_MARKERS"
AUTOJINJA_SILENT         = "AUTOJINJA_SILENT"
AUTOJINJA_THIS_DIRPATH   = "THIS_DIRPATH"



import os
import traceback

### Base exception

class CommonException(Exception):
    @staticmethod
    def from_marker(exceptionType, marker, pos, size, message):
        line = line_at_index(marker.string, pos)
        (l, c) = index_to_coordinates(marker.string, pos)
        message = f"{message}\n{line}\n{' ' * (c-1)}{'^' * size} line {l}, column {c}"
        exception = exceptionType(message)
        exception.coordinates = (l, c)
        return exception
    @staticmethod
    def from_exception(exception, marker):
        exception = prepend_jinja2_traceback(exception)
        text = format_text(marker.header)
        (l, c) = index_to_coordinates(marker.string, marker.header_open)
        stack = f"\n  During {'reinsertion' if marker.is_edit else 'generation'} of \"{marker.open} {text} {marker.close}\" at line {l}, column {c}"
        message = str(exception).lstrip('\n')
        message = f"{stack}\n{message}"
        exception = wrap_exception(exception, message)
        exception.coordinates = (l, c)
        return exception

    def __init__(self, message):
        super().__init__(message)

### Parsing exception

class ParsingException(CommonException):
    def __init__(self, message):
        super().__init__(message)

class OpenMarkerNotFoundException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(OpenMarkerNotFoundException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Couldn't find corresponding open marker \"{marker.open} {marker.close}\":")
    def __init__(self, message):
        super().__init__(message)

class CloseMarkerNotFoundException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(CloseMarkerNotFoundException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Couldn't find corresponding close marker \"{marker.close}\":")
    def __init__(self, message):
        super().__init__(message)

class EndMarkerNotFoundException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(EndMarkerNotFoundException,
                                           marker,
                                           marker.header_close - len(marker.close),
                                           len(marker.close),
                                           f"Couldn't find corresponding end marker \"{marker.open} {marker.end} {marker.close}\":")
    def __init__(self, message):
        super().__init__(message)

class RequireHeaderInlineException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(RequireHeaderInlineException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Marker can't have a multiline header:")
    def __init__(self, message):
        super().__init__(message)

class RequireHeaderMultilineException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(RequireHeaderMultilineException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Marker can't have a one line header:")
    def __init__(self, message):
        super().__init__(message)

class WrongHeaderIndentationException(ParsingException):
    @staticmethod
    def from_marker(marker, pos):
        return CommonException.from_marker(WrongHeaderIndentationException,
                                           marker,
                                           pos,
                                           len(marker.open),
                                           f"Wrong marker header indentation:")
    def __init__(self, message):
        super().__init__(message)

class RequireNewlineException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(RequireNewlineException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Marker can't start on same line as previous end marker:")
    def __init__(self, message):
        super().__init__(message)

class RequireInlineException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(RequireInlineException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Marker must start on same line as previous marker:")
    def __init__(self, message):
        super().__init__(message)

class WrongInclusionException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(WrongInclusionException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Marker has wrong inclusion regarding enclosing markers:")
    def __init__(self, message):
        super().__init__(message)

class DuplicateEditException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(DuplicateEditException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Duplicate edit marker \"{marker.open} {marker.header} {marker.close}\", consider reusing/removing duplicates:")
    def __init__(self, message):
        super().__init__(message)

class DirectlyEnclosedEditException(ParsingException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(DirectlyEnclosedEditException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Directly enclosed edit marker \"{marker.open} {marker.header} {marker.close}\", consider reusing/removing it:")
    def __init__(self, message):
        super().__init__(message)

### Generation exception

class GenerationException(CommonException):
    def __init__(self, message):
        super().__init__(message)

class RequireBodyInlineException(GenerationException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(RequireBodyInlineException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Generated body must contain only one line to be inlined:")
    def __init__(self, message):
        super().__init__(message)

class NonGeneratedEditException(GenerationException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(NonGeneratedEditException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Non-generated edit marker \"{marker.open} {marker.header} {marker.close}\", consider reusing/removing it:")
    def __init__(self, message):
        super().__init__(message)

class AlreadyGeneratedEditException(GenerationException):
    @staticmethod
    def from_marker(marker):
        return CommonException.from_marker(AlreadyGeneratedEditException,
                                           marker,
                                           marker.header_open,
                                           len(marker.open),
                                           f"Already generated edit marker \"{marker.open} {marker.header} {marker.close}\", consider fixing generation:")
    def __init__(self, message):
        super().__init__(message)

### Utils

def index_to_coordinates(string, index):
    """ Returns the corresponding tuple (line, column) of the character at the given index of the given string.
    """
    if index < 0:
        index =  index % len(string)
    sp = string[:index+1].splitlines(keepends=True)
    return len(sp), len(sp[-1])

def line_at_index(string, index):
    """ Returns the line at the given index of the given string.
    """
    if index < 0:
        index =  index % len(string)
    start = string.rfind('\n', 0, index)+1
    end = string.find('\n', index)
    suffix = '\\n' if end >= 0 else '\\0'
    if end < 0:
        end = len(string)
    return f"{string[start:end]}{suffix}"

def format_text(text, max_length = 50, suffix = "..."):
    """ Replaces tabulations and ends of line.
    """
    text = text if len(text) <= max_length else f"{text[:max_length]}{suffix}"
    return text.replace('\t', "\\t").replace('\n', "\\n")

def split_traceback(tb, remove_class_name = False):
    """ Splits the given traceback into (stacktrace, message).
    """
    startidx = -1
    while True:
        startidx = tb.find('\n', startidx+1)
        if startidx < 0:
            break # No stacktrace
        if not tb.startswith('  ', startidx+1):
            break # End of stacktrace
    if startidx < 0:
        stacktrace = None # No stacktrace
    else:
        stacktrace = tb[:startidx]
    if not remove_class_name:
        message = tb[startidx+1:]
    else:
        idx = tb.find(' ', startidx+1)
        if idx < 0:
            message = None # No message
        else:
            idx += 2 if tb[idx+1] == '\n' else 1 # lstrip one
            message = tb[idx:]
    return stacktrace, message

def wrap_exception(exception, message):
    """ Wraps the given exception with a custom message.
        Creates an object that derives the exception class.
    """
    if hasattr(exception, "__exception") and hasattr(exception, "__message"):
        exception.__message = message
        return exception
    def __init__(self):
        pass
    def __repr__(self):
        return self.__message
    def __str__(self):
        return self.__message
    dicts = { "__init__" : __init__,
              "__repr__" : __repr__,
              "__str__" : __str__,
              "__module__" : exception.__class__.__module__ }
    wrapped_exception = type(exception.__class__.__qualname__, (exception.__class__,), dicts)()
    wrapped_exception.__dict__.update(exception.__dict__)
    wrapped_exception.__exception = exception
    wrapped_exception.__message = message
    return wrapped_exception

def prepend_jinja2_traceback(exception, tb = None):
    """ Prepends the Jinja2 traceback to the given exception.
        Creates an object that derives the exception class.
    """
    if tb == None:
        tb = traceback.format_exc().rstrip('\n')
    (stacktrace, message) = split_traceback(tb, True)
    if stacktrace != None:
        token = f"jinja2{os.sep}environment.py\", " # Jinja2 hack
        startidx = -1
        while True:
            idx = stacktrace.find(token, startidx+1)
            if idx < 0:
                break # Token not found
            startidx = stacktrace.find('\n', idx+len(token))
            if startidx < 0:
                startidx = len(stacktrace)
                break # No eol after token
            if not stacktrace.startswith('    ', startidx+1):
                continue # No inner traceback
            startidx = stacktrace.find('\n', startidx+1)
            if startidx < 0:
                startidx = len(stacktrace)
                break # No eol after inner traceback
        if startidx < 0:
            return exception
        stacktrace = stacktrace[startidx+1:]
        if len(stacktrace) == 0:
            stacktrace = None
        else:
            # Clean autojinja stacktrace
            token = f"autojinja{os.sep}templates.py\", "
            idx = stacktrace.find(token)
            if idx >= 0:
                startidx = stacktrace.rfind('\n', 0, idx)
                if startidx < 0: # No eol before token
                    stacktrace = None
                else:
                    stacktrace = stacktrace[:startidx]
    if message == None:
        message = ""
    message = f"\n{stacktrace}\n{message}" if stacktrace else f"\n{message}"
    return wrap_exception(exception, message)

def clean_traceback(exception):
    """ Removes the traceback of the given exception.
    """
    return prepend_jinja2_traceback(exception).with_traceback(None)
